@@ -1,10 +1,12 @@
 import ast
 import asyncio
 import dataclasses
+import datetime
 import logging
 import re
+from contextlib import suppress
 from copy import deepcopy
-from typing import Any, List, Optional, Tuple
+from typing import Any, Callable
 
 import jq
 import pandas as pd
@@ -29,11 +31,18 @@ RE_SET_KEEP_TYPE = r'{{__keep_type__\1}}\2'
 RE_GET_KEEP_TYPE = r'{{(__keep_type__[^({{)}]*)}}'
 RE_NAMED_PARAM = r'\'?%\([a-zA-Z0-9_]*\)s\'?'
 
+# If a parameter has one of these values, it'll be removed from a query
+_PARAMETER_VALUES_TO_ELIMINATE = ('__VOID__',)
+
 
 class NonValidVariable(Exception):
     """Error thrown for a non valid variable in endpoint"""
 
 
+class ClusterStartException(Exception):
+    """Raised when start cluster fails"""
+
+
 def is_jinja_alone(s: str) -> bool:
     """
     Return True if the given string is a jinja template alone.
@@ -52,119 +61,138 @@ def is_jinja_alone(s: str) -> bool:
     return re.match(RE_JINJA_ALONE, s) or (s.startswith('{%') and s.endswith('%}'))
 
 
-def nosql_apply_parameters_to_query(query, parameters, handle_errors=False):
+def _has_parameters(query: dict | list[dict] | tuple | str) -> bool:
+    t = Environment().parse(query)
+    return bool(meta.find_undeclared_variables(t) or re.search(RE_PARAM, query))
+
+
+def _prepare_parameters(p: dict | list[dict] | tuple | str) -> dict | list[dict] | tuple | str:
+    if isinstance(p, str):
+        return repr(p)
+    elif isinstance(p, list):
+        return [_prepare_parameters(e) for e in p]
+    elif isinstance(p, dict):
+        return {k: _prepare_parameters(v) for k, v in p.items()}
+    else:
+        return p
+
+
+def _prepare_result(res: dict | list[dict] | tuple | str) -> dict | list[dict] | tuple | str:
+    if isinstance(res, str):
+        return ast.literal_eval(res)
+    elif isinstance(res, list):
+        return [_prepare_result(e) for e in res]
+    elif isinstance(res, dict):
+        return {k: _prepare_result(v) for k, v in res.items()}
+    else:
+        return res
+
+
+def _flatten_rendered_nested_list(origin: list, rendered: list) -> list:
+    """
+    Flatten rendered lists in the parent list, so we have the same render logic
+    as in toucan frontend's templates.
+    """
+    result = []
+    for elem, rendered_elem in zip(origin, rendered):
+        if isinstance(elem, str) and isinstance(rendered_elem, list):
+            # a list has been rendered: flatten the result
+            result += rendered_elem
+        else:
+            result.append(rendered_elem)
+    return result
+
+
+def _render_query(query: dict | list[dict] | tuple | str, parameters: dict):
+    """
+    Render both jinja or %()s templates in query
+    while keeping type of parameters
+    """
+    if isinstance(query, dict):
+        return {key: _render_query(value, parameters) for key, value in deepcopy(query).items()}
+    elif isinstance(query, list):
+        rendered_query = [_render_query(elt, parameters) for elt in deepcopy(query)]
+        rendered_query = _flatten_rendered_nested_list(query, rendered_query)
+        return rendered_query
+    elif isinstance(query, tuple):
+        return tuple(_render_query(value, parameters) for value in deepcopy(query))
+    elif isinstance(query, str):
+        if not _has_parameters(query):
+            return query
+
+        # Replace param templating with jinja templating:
+        query = re.sub(RE_PARAM, r'{{ \g<1> }}', query)
+
+        # Add quotes to string parameters to keep type if not complex
+        clean_p = deepcopy(parameters)
+        if re.match(RE_JINJA_ALONE, query):
+            clean_p = _prepare_parameters(clean_p)
+
+        if is_jinja_alone(query):
+            env = NativeEnvironment()
+        else:
+            env = Environment()
+
+        res = env.from_string(query).render(clean_p)
+        # NativeEnvironment's render() isn't recursive, so we need to
+        # apply recursively the literal_eval by hand for lists and dicts:
+        if isinstance(res, (list, dict)):
+            return _prepare_result(res)
+        return res
+    else:
+        return query
+
+
+def _handle_missing_params(elt: dict | list[dict] | tuple | str, params: dict, handle_errors: bool):
+    """
+    Remove a dictionary key if its value has a missing parameter.
+    This is used to support the __VOID__ syntax, specific at Toucan Toco :
+        cf. https://bit.ly/2Ln6rcf
+    """
+    if isinstance(elt, dict):
+        e = {}
+        for k, v in elt.items():
+            if isinstance(v, str):
+                missing_params = []
+                # In case the query was interpolated before being passed to the connector
+                if v.strip() in _PARAMETER_VALUES_TO_ELIMINATE:
+                    continue
+                matches = re.findall(RE_PARAM, v) + re.findall(RE_JINJA, v)
+
+                for m in matches:
+                    try:
+                        rendered = Template('{{ %s }}' % m, undefined=StrictUndefined).render(
+                            params
+                        )
+                        if rendered in _PARAMETER_VALUES_TO_ELIMINATE:
+                            missing_params.append(m)
+                    except Exception:
+                        if handle_errors:
+                            raise NonValidVariable(f'Non valid variable {m}')
+                        missing_params.append(m)
+                if any(missing_params):
+                    continue
+                else:
+                    e[k] = v
+            else:
+                e[k] = _handle_missing_params(v, params, handle_errors)
+        return e
+    elif isinstance(elt, list):
+        # Not filtering out empty dicts here because they may have a meaning in some backends
+        return [_handle_missing_params(e, params, handle_errors) for e in elt]
+    else:
+        return elt
+
+
+def nosql_apply_parameters_to_query(
+    query: dict | list[dict] | tuple | str, parameters: dict | None, handle_errors: bool = False
+):
     """
     WARNING : DO NOT USE THIS WITH VARIANTS OF SQL
     Instead use your client library parameter substitution method.
     https://www.owasp.org/index.php/Query_Parameterization_Cheat_Sheet
     """
 
-    def _has_parameters(query):
-        t = Environment().parse(query)
-        return meta.find_undeclared_variables(t) or re.search(RE_PARAM, query)
-
-    def _prepare_parameters(p):
-        if isinstance(p, str):
-            return repr(p)
-        elif isinstance(p, list):
-            return [_prepare_parameters(e) for e in p]
-        elif isinstance(p, dict):
-            return {k: _prepare_parameters(v) for k, v in p.items()}
-        else:
-            return p
-
-    def _prepare_result(res):
-        if isinstance(res, str):
-            return ast.literal_eval(res)
-        elif isinstance(res, list):
-            return [_prepare_result(e) for e in res]
-        elif isinstance(res, dict):
-            return {k: _prepare_result(v) for k, v in res.items()}
-        else:
-            return res
-
-    def _flatten_rendered_nested_list(origin: list, rendered: list) -> list:
-        """
-        Flatten rendered lists in the parent list, so we have the same render logic
-        as in toucan frontend's templates.
-        """
-        result = []
-        for elem, rendered_elem in zip(origin, rendered):
-            if isinstance(elem, str) and isinstance(rendered_elem, list):
-                # a list has been rendered: flatten the result
-                result += rendered_elem
-            else:
-                result.append(rendered_elem)
-        return result
-
-    def _render_query(query, parameters):
-        """
-        Render both jinja or %()s templates in query
-        while keeping type of parameters
-        """
-        if isinstance(query, dict):
-            return {key: _render_query(value, parameters) for key, value in deepcopy(query).items()}
-        elif isinstance(query, list):
-            rendered_query = [_render_query(elt, parameters) for elt in deepcopy(query)]
-            rendered_query = _flatten_rendered_nested_list(query, rendered_query)
-            return rendered_query
-        elif isinstance(query, str):
-            if not _has_parameters(query):
-                return query
-
-            # Replace param templating with jinja templating:
-            query = re.sub(RE_PARAM, r'{{ \g<1> }}', query)
-
-            # Add quotes to string parameters to keep type if not complex
-            clean_p = deepcopy(parameters)
-            if re.match(RE_JINJA_ALONE, query):
-                clean_p = _prepare_parameters(clean_p)
-
-            if is_jinja_alone(query):
-                env = NativeEnvironment()
-            else:
-                env = Environment()
-
-            res = env.from_string(query).render(clean_p)
-            # NativeEnvironment's render() isn't recursive, so we need to
-            # apply recursively the literal_eval by hand for lists and dicts:
-            if isinstance(res, (list, dict)):
-                return _prepare_result(res)
-            return res
-        else:
-            return query
-
-    def _handle_missing_params(elt, params, handle_errors):
-        """
-        Remove a dictionary key if its value has a missing parameter.
-        This is used to support the __VOID__ syntax, specific at Toucan Toco :
-            cf. https://bit.ly/2Ln6rcf
-        """
-        if isinstance(elt, dict):
-            e = {}
-            for k, v in elt.items():
-                if isinstance(v, str):
-                    matches = re.findall(RE_PARAM, v) + re.findall(RE_JINJA, v)
-                    missing_params = []
-                    for m in matches:
-                        try:
-                            Template('{{ %s }}' % m, undefined=StrictUndefined).render(params)
-                        except Exception:
-                            if handle_errors:
-                                raise NonValidVariable(f'Non valid variable {m}')
-                            missing_params.append(m)
-                    if any(missing_params):
-                        continue
-                    else:
-                        e[k] = v
-                else:
-                    e[k] = _handle_missing_params(v, params, handle_errors)
-            return e
-        elif isinstance(elt, list):
-            return [_handle_missing_params(e, params, handle_errors) for e in elt]
-        else:
-            return elt
-
     query = _handle_missing_params(query, parameters, handle_errors)
 
     if parameters is None:
@@ -277,10 +305,10 @@ class HttpError(Exception):
 
 @dataclasses.dataclass()
 class ConnectorStatus:
-    status: Optional[bool] = None
-    message: Optional[str] = None
-    error: Optional[str] = None
-    details: Optional[List[Tuple[str, Optional[bool]]]] = dataclasses.field(default_factory=list)
+    status: bool | None = None
+    message: str | None = None
+    error: str | None = None
+    details: list[tuple[str, bool | None]] | None = dataclasses.field(default_factory=list)
 
     def to_dict(self):
         return dataclasses.asdict(self)
@@ -294,9 +322,7 @@ def get_param_name(printf_style_argument: str) -> str:
     return printf_style_argument[2:-2]
 
 
-def convert_to_qmark_paramstyle(
-    query_string: str, params_values: dict
-) -> Tuple[str, List[Optional[Any]]]:
+def convert_to_qmark_paramstyle(query_string: str, params_values: dict) -> tuple[str, list[Any]]:
     """Takes a query in pyformat paramstyle and transforms it in qmark
        by replacing placeholders by ? and returning values in right order
     ex :
@@ -356,6 +382,25 @@ def is_interpolating_table_name(query: str) -> bool:
     return table_name.startswith('%(')
 
 
+def infer_datetime_dtype(df: pd.DataFrame) -> None:
+    """
+    Even if a RDBMS table's column has type `date NOT NULL`,
+    we get a `object` dtype in the resulting pandas dataframe.
+    This util allows to automatically convert it to `datetime64[ns]`.
+    """
+    for colname in df:
+        if df[colname].dtype == 'object':
+            # get the first non-null value in the series.
+            # if it's a datetime, try to convert the whole serie to datetime dtype.
+            s = df[colname]
+            idx = s.first_valid_index()
+            if idx is not None:
+                first_value = s.loc[idx]
+                if isinstance(first_value, (datetime.datetime, datetime.date)):
+                    with suppress(Exception):
+                        df[colname] = pd.to_datetime(df[colname], errors='coerce')
+
+
 def pandas_read_sql(
     query: str,
     con,
@@ -376,6 +421,11 @@ def pandas_read_sql(
         params = adapt_param_type(params)
 
     try:
+        # FIXME: We should use here the sqlalchemy.text() module to
+        # escape characters like % but as a quick fix,
+        # we added regex replace that will exclude %(.*) compositions
+        query = query.replace('%%', '%')
+        query = re.sub(r'%[^(%]', r'%\g<0>', query)
         df = pd.read_sql(query, con=con, params=params, **kwargs)
     except pd.io.sql.DatabaseError:
         if is_interpolating_table_name(query):
@@ -384,4 +434,24 @@ def pandas_read_sql(
         else:
             raise
 
+    infer_datetime_dtype(df)
     return df
+
+
+def sanitize_query(
+    query: str, params: dict[str, object], transformer: Callable[[str], str]
+) -> tuple[str, dict[str, object]]:
+    """
+    We allow jinja templates in queries but we don't want to interpolate them directly to avoid SQL injections
+    So we extract the jinja templates and replace them with placeholders and interpolate the placeholders separately.
+    We then send the query with placeholders and the interpolated values
+    to the SQL driver that will reject or not the query!
+    """
+    import re
+
+    all_query_params = re.findall(r'{{.*?}}', query)
+    for i, query_param in enumerate(all_query_params):
+        params[f'__QUERY_PARAM_{i}__'] = nosql_apply_parameters_to_query(query_param, params)
+        query = query.replace(query_param, transformer(f'__QUERY_PARAM_{i}__'))
+
+    return query, params
